import json
import numpy as np
from typing import List
import math
import cv2

coco_limbSeq = [
    [2, 3], [2, 6], [3, 4], [4, 5], 
    [6, 7], [7, 8], [2, 9], [9, 10], 
    [10, 11], [2, 12], [12, 13], [13, 14], 
    [2, 1], [1, 15], [15, 17], [1, 16], 
    [16, 18],
]
coco_colors = [
    [255, 0, 0], [255, 85, 0], [255, 170, 0], [255, 255, 0], [170, 255, 0], [85, 255, 0], [0, 255, 0],
    [0, 255, 85], [0, 255, 170], [0, 255, 255], [0, 170, 255], [0, 85, 255], [0, 0, 255], [85, 0, 255],
    [170, 0, 255], [255, 0, 255], [255, 0, 170], [255, 0, 85]
]

body_25_limbSeq = [
    [1, 8], [1, 2], [1, 5], [2, 3], 
    [3, 4], [5, 6], [6, 7], [8, 9], 
    [9, 10], [10, 11], [8, 12], [12, 13], 
    [13, 14], [1, 0], [0, 15], [15, 17], 
    [0, 16], [16, 18], [14, 19], [19, 20], 
    [14, 21], [11, 22], [22, 23], [11, 24]
]
body_25_colors = [
    [255, 0, 85], [255, 0, 0], [255, 85, 0], [255, 170, 0], [255, 255, 0], [170, 255, 0], [85, 255, 0],
    [0, 255, 0], [255, 0, 0], [0, 255, 85], [0, 255, 170], [0, 255, 255], [0, 170, 255], [0, 85, 255],
    [0, 0, 255], [255, 0, 170], [170, 0, 255], [255, 0, 255], [85, 0, 255], [0, 0, 255], [0, 0, 255],
    [0, 0, 255], [0, 0, 255], [0, 255, 255], [0, 255, 255], [0, 255, 255]
]

body_25B_limbSeq = [
    [0, 1], [0, 2], [1, 3], [2, 4], [5, 7], [6, 8], [7, 9], [8, 10], 
    [5, 11], [6, 12], [11, 13], [12, 14], [13, 15], [14, 16], [15, 19], 
    [19, 20], [15, 21], [16, 22], [22, 23], [16, 24], [5, 17], [6, 17], 
    [17, 18], [11, 12]
]

body_25B_colors = [
    [255, 0, 85], [170, 0, 255], [255, 0, 170], [85, 0, 255], [255, 0, 255], 
    [170, 255, 0], [255, 85, 0], [85, 255, 0], [255, 170, 0], [0, 255, 0], 
    [255, 255, 0], [0, 170, 255], [0, 255, 85], [0, 85, 255], [0, 255, 170], 
    [0, 0, 255], [0, 255, 255], [255, 0, 0], [255, 0, 0], [0, 0, 255], 
    [0, 0, 255], [0, 0, 255], [0, 255, 255], [0, 255, 255], [0, 255, 255]
]

class Keypoint:
    def __init__(self, x: float, y: float, confidence: float = 1.0):
        """
        Initialize a Keypoint object.

        Args:
            x (float): The x-coordinate of the keypoint.
            y (float): The y-coordinate of the keypoint.
            confidence (float): The confidence score of the keypoint. Default is 1.0.
        """
        self.x = x
        self.y = y
        self.confidence = confidence

    def __repr__(self):
        return f"Keypoint(x={self.x}, y={self.y}, confidence={self.confidence})"

class Skeleton:
    def __init__(self, keypoints: List[Keypoint]):
        self.keypoints = keypoints

    def __repr__(self):
        return f"Skeleton(keypoints={self.keypoints})"

class Skeleton_Seqence:
    def __init__(self, skeletons: List[Skeleton]):
        self.skeletons = skeletons

    def __repr__(self):
        return f"Skeleton_Seqence(Skeleton_frames={self.skeletons})"
    
    def get_frame(self, frame_index: int) -> Skeleton:
        return self.skeletons[frame_index]
    
    def add_frame(self, skeleton: Skeleton):
        self.skeletons.append(skeleton)
    
def get_time_slice_for_Skeleton_Seqences(skeleton_seqences: List[Skeleton_Seqence], frame_index: int) -> List[Skeleton]:
    return [skeleton_seq.get_frame(frame_index) for skeleton_seq in skeleton_seqences]
        

def is_normalized(keypoints: List[Keypoint]) -> bool:
    for keypoint in keypoints:
        if not (0 <= keypoint.x <= 1 and 0 <= keypoint.y <= 1):
            return False
    return True

def draw_bodypose(canvas: np.ndarray, keypoints: List[Keypoint], limbSeq, colors, xinsr_stick_scaling: bool = False) -> np.ndarray:
    """
    Draw keypoints and limbs representing body pose on a given canvas.

    Args:
        canvas (np.ndarray): A 3D numpy array representing the canvas (image) on which to draw the body pose.
        keypoints (List[Keypoint]): A list of Keypoint objects representing the body keypoints to be drawn.
        xinsr_stick_scaling (bool): Whether or not scaling stick width for xinsr ControlNet

    Returns:
        np.ndarray: A 3D numpy array representing the modified canvas with the drawn body pose.

    Note:
        The function expects the x and y coordinates of the keypoints to be normalized between 0 and 1.
    """
    if not is_normalized(keypoints):
        H, W = 1.0, 1.0
    else:
        H, W, _ = canvas.shape

    CH, CW, _ = canvas.shape
    stickwidth = 4

    # Ref: https://huggingface.co/xinsir/controlnet-openpose-sdxl-1.0
    max_side = max(CW, CH)
    if xinsr_stick_scaling:
        stick_scale = 1 if max_side < 500 else min(2 + (max_side // 1000), 7)
    else:
        stick_scale = 1

    for (k1_index, k2_index), color in zip(limbSeq, colors):
        keypoint1 = keypoints[k1_index - 1]
        keypoint2 = keypoints[k2_index - 1]

        if keypoint1 is None or keypoint2 is None:
            continue

        Y = np.array([keypoint1.x, keypoint2.x]) * float(W)
        X = np.array([keypoint1.y, keypoint2.y]) * float(H)
        mX = np.mean(X)
        mY = np.mean(Y)
        length = ((X[0] - X[1]) ** 2 + (Y[0] - Y[1]) ** 2) ** 0.5
        angle = math.degrees(math.atan2(X[0] - X[1], Y[0] - Y[1]))
        polygon = cv2.ellipse2Poly((int(mY), int(mX)), (int(length / 2), stickwidth*stick_scale), int(angle), 0, 360, 1)
        cv2.fillConvexPoly(canvas, polygon, [int(float(c) * 0.6) for c in color])

    for keypoint, color in zip(keypoints, colors):
        if keypoint is None:
            continue

        x, y = keypoint.x, keypoint.y
        x = int(x * W)
        y = int(y * H)
        cv2.circle(canvas, (int(x), int(y)), 4, color, thickness=-1)

    return canvas

def json_to_keypoints_openpose(json_file: str) -> List[Keypoint]:
    with open(json_file, 'r') as file:
        data = json.load(file)
        keypoints = data[0]['people'][0]['pose_keypoints_2d']
        keypoints = [Keypoint(keypoints[i], keypoints[i + 1]) for i in range(0, len(keypoints), 3)]
    return keypoints

def coordinates_to_keypoints(coordinates: list) -> List[Keypoint]:
    keypoints = [Keypoint(coordinates[i], coordinates[i + 1]) for i in range(0, len(coordinates), 3)]
    return keypoints

def save_bodypose(width: int, height: int, coordinates: list):
    if not hasattr(save_bodypose, 'counter'):
        save_bodypose.counter = 0  # Initialize the counter attribute

    canvas = np.zeros((height, width, 3), dtype=np.uint8)
    keypoints = coordinates_to_keypoints(coordinates)
    canvas = draw_bodypose(canvas, keypoints, coco_limbSeq, coco_colors)

    # Save as body_pose_output0000.png, body_pose_output0001.png, ...
    cv2.imwrite('body_pose_output%04d.png' % save_bodypose.counter, canvas)
    save_bodypose.counter += 1  # Increment the counter

def array_json_to_Skeleton_Seqences(json_file: str) -> List[Skeleton_Seqence]:
    with open(json_file, 'r') as file:
        data = json.load(file)
    
    skeleton_sequences = []
    for frame in data:
        for i in range(len(frame)):
            while len(skeleton_sequences) <= i:
                skeleton_sequences.append(None)
                skeleton_sequences[i] = Skeleton_Seqence([])
            skeleton = Skeleton([Keypoint(keypoint[0], keypoint[1], keypoint[2]) for keypoint in frame[i]])
            skeleton_sequences[i].add_frame(skeleton)
    return skeleton_sequences

def main():
    json_file = '0001_002_00_01_1.json'
    image_path = 'test'
    skeleton_sequences = array_json_to_Skeleton_Seqences(json_file)
    for i in range(20):
        sliced = get_time_slice_for_Skeleton_Seqences(skeleton_sequences, i)
        canvas = np.zeros((480, 640, 3), dtype=np.uint8)
        for skeleton in sliced:
            keypoints = skeleton.keypoints
            canvas = draw_bodypose(canvas, keypoints, body_25B_limbSeq, body_25B_colors)

        cv2.imwrite(image_path + '_' + str(i) + '.png', canvas)

if __name__ == '__main__':
    main()